C This subroutine extracts the roots of the polynome a2*z^2+a1*z+a0=0
C with the respective values zm and zp for the -/+ depending on the sign
C of the squareroot. The coefficients 'a' are real and z's are complex
C zdiff=zp-zm
C This file is part of aITALC
C Copyright (C) 2004 Alejandro Lorca <alejandro.lorca@desy.de>
C-----------------------------------------------------------------------
      SUBROUTINE DGETROOTS(a2,a1,a0,zm,zp,zdiff,index)
      implicit none
      double precision a2,a1,a0
      double complex zm,zp,zdiff,ztrans
      integer index
      double precision term1,term2,expan,theroot,x,ra0,ra1
      double precision DSIGN
      external DSIGN
C-----------------------------------------------------------------------
c      print *,'Coeff a0=',a0
c      print *,'Coeff a1=',a1
c      print *,'Coeff a2=',a2
C-----------------------------------------------------------------------
C Trivial case with a2=0
      if (a2.EQ.0d0) then
        print *,'dgetroots: error, coefficient a2 of z^2 = 0'
        stop
      endif
C-----------------------------------------------------------------------
C Trivial case with a0=0
      if (a0.EQ.0d0) then
        if (a1.EQ.0d0) then
c          print *,'a0=a1=0'
          zp=DCMPLX(0)
          zm=zp
          zdiff=zp
          index=0
          goto 20
        endif
        zp=DCMPLX(-a1/a2)
        zm=DCMPLX(0)
        if (-a1/a2.LT.0d0) then
          zm=zp
          zp=DCMPLX(0)
          zdiff=zm
        endif
c        print *,'a0=0'
        index=1
        goto 20
      endif
C-----------------------------------------------------------------------
C Trivial case with a1=0

      if ((a1.EQ.0d0).AND.(-a0/a2.GT.0d0)) then
        zp=+DCMPLX(DSQRT(-a0/a2))
        zm=-zp
        zdiff=2*zp
        index=1
        goto 20
      elseif ((a1.EQ.0d0).AND.(-a0/a2.LT.0d0)) then
        zp=+DCMPLX(0d0,DSQRT(a0/a2))
        zm=-zp
        zdiff=2*zp
        index=-1
        goto 20
      endif
C-----------------------------------------------------------------------
C Non-trivial case.
      term1=a1**2
      term2=4*a2*a0
      x=term2/term1
      expan=x/2*(
     &         +1
     &         +x/4
     &         +x**2/8
     &         )
      if (DABS(1-x).LT.2d-15) then
c        print *,'the same roots'
C Sqrt=0, same roots, within 2d-15 security round-offset
        zp=DCMPLX(-a1/(2*a2))
        zm=zp
        zdiff=DCMPLX(0)
        index=0
        goto 20

      elseif (term1.GT.term2) then
c        print *,'real roots'
C Real roots
        theroot=DSQRT(1-x)
        if ((DABS(x).LT.3d-4).AND.(DSIGN(a1).GT.0d0)) then
c          print *,'expansion +'
C Better to use expansion
          zp=DCMPLX(-expan)
          zm=DCMPLX(-1-theroot)
        elseif ((DABS(x).LT.3d-4).AND.(DSIGN(a1).LT.0d0)) then
c          print *,'expansion -'
          zp=DCMPLX(+1+theroot)
          zm=DCMPLX(+expan)
        else
c          print *,'no expansion'
          zp=DCMPLX(-DSIGN(a1)+theroot)
          zm=DCMPLX(-DSIGN(a1)-theroot)
        endif
        zp=zp*DCMPLX(DABS(a1)/(2*a2))
        zm=zm*DCMPLX(DABS(a1)/(2*a2))
        zdiff=DCMPLX(DABS(a1)*theroot/a2)
        index=1
c        print *,'rec a0=',+a2*dble(zp*zm)
c        print *,'rec a1=',-a2*dble(zp+zm)
        goto 20
      elseif (term1.LT.term2) then
c        print *,'complex roots'
C Complex roots
        theroot=DSQRT(-(1-x))
        zp=DCMPLX(-DSIGN(a1),+theroot)
        zm=DCMPLX(-DSIGN(a1),-theroot)
        zp=zp*DCMPLX(DABS(a1)/(2*a2))
        zm=zm*DCMPLX(DABS(a1)/(2*a2))
        zdiff=DCMPLX(0d0,DABS(a1)*theroot/a2)
        index=-1
        goto 20
      endif
C-----------------------------------------------------------------------
 20   continue
C Ordering; the zm is expected to be lower also if sign(c2) < 0
      if (DBLE(zp).LT.DBLE(zm)) then
C Do the change,
        ztrans=zp
        zp=zm
        zm=ztrans
      endif
C Reconstruction test
      ra0=+a2*DBLE(zp*zm)
      ra1=-a2*DBLE(zp+zm)
      if ((a0.EQ.0d0).AND.(ra0.NE.0d0)) then
        print *,'dgetroots: error on a0 reconstruction'
     &       ,DABS(ra0-a0)
        print *,'a0=',a0,ra0   
        goto 30
      endif
      if ((a1.EQ.0d0).AND.(ra1.NE.0d0)) then
        print *,'dgetroots: error on a1 reconstruction'
     &       ,DABS(ra1-a1)
        print *,'a1=',a1,ra1
        goto 30
      endif
      if (DABS((ra0-a0)/a0).GT.3D-12) then
        print *,'dgetroots: error on a0 reconstruction'
     &       ,DABS((ra0-a0)/a0)
        print *,'a0=',a0,ra0
        stop
      elseif (DABS((ra1-a1)/a1).GT.3D-12) then
        print *,'dgetroots: error on a1 reconstruction'
     &       ,DABS((ra1-a1)/a1)
        print *,'a1=',a1,ra1
        stop
      endif
C-----------------------------------------------------------------------
 30   return
      end

C $Id: dgetroots.f,v 1.4 2004/10/07 15:05:11 alorca Exp $
